import torch
import argparse
import cv2
import numpy as np
import os
import utils.custom_transforms as transformer
from utils.patch_transformer import Adversarial_Patch
from models.adversarial_models import AdversarialModels
from utils.utils import makedirs, to_cuda_vars
from utils.dataloader import SingleImageLoader
import matplotlib.pyplot as plt
from tqdm import tqdm
import warnings
warnings.simplefilter('ignore')

parser = argparse.ArgumentParser(description='Generating Adversarial Patches')
parser.add_argument('--img_path', type=str, default='Src/input_img/test_scene1.jpg')
parser.add_argument('--distill_ckpt', type=str, default="models/guo/distill_model.ckpt")
parser.add_argument('--height', type=int, help='input image height', default=256)
parser.add_argument('--width', type=int, help='input image width', default=512)
parser.add_argument('--seed', type=int, help='seed for random functions, and network initialization', default=0)
parser.add_argument('--patch_size', type=int, help='Resolution of patch', default=256)
parser.add_argument('--patch_shape', type=str, help='circle or square', default='circle')
parser.add_argument('--patch_path', type=str, default='Src/trained_patches/circle_patch_distill_near.npy')
parser.add_argument('--mask_path', type=str, default='Src/trained_patches/circle_mask.npy')
parser.add_argument('--no_overwrite', help='', action='store_true')
parser.add_argument('--model', nargs='*', type=str, default='distill', choices=['distill'], help='Model architecture')
parser.add_argument('--name', type=str, help='output directory', default='')
args = parser.parse_args()


def main():
    if args.name:
        save_path = os.path.join('Dst/test_result', args.name)
    else:
        name = (os.path.splitext(args.patch_path)[0]).replace('/', '_')
        save_path = os.path.join('Dst/test_result', name)
    print('===============================')
    print('=> Everything will be saved to \"{}\"'.format(save_path))
    makedirs(save_path)

    torch.manual_seed(args.seed)
    torch.backends.cudnn.benchmark = True

    # DataLoader
    test_transform = transformer.Compose([
        transformer.ResizeImage(h=args.height, w=args.width),
        transformer.ArrayToTensor()
    ])

    test_set = SingleImageLoader(
        img_path=args.img_path,
        seed=args.seed,
        transform=test_transform
    )

    test_loader = torch.utils.data.DataLoader(dataset=test_set, batch_size=1)

    print('===============================')
    # Attacked Models
    models = AdversarialModels(args)
    models.load_ckpt()

    # Patch and Mask
    adv = Adversarial_Patch(
        patch_type=args.patch_shape,
        batch_size=1,
        image_size=min(args.width, args.height),
        patch_size=args.patch_size,
        train=False
    )

    patch_cpu, mask_cpu = adv.load_patch_and_mask_from_file(args.patch_path, args.mask_path, npy=True)

    print('===============================')
    for i_batch, sample in tqdm(enumerate(test_loader), total=len(test_loader)):
        sample = to_cuda_vars(sample)  # send item to gpu
        sample.update(models.get_original_disp(to_cuda_vars(sample)))  # get non-attacked disparity

        img, original_disp = sample['left'], sample['original_distill_disp']
        patch, mask = patch_cpu.cuda(), mask_cpu.cuda()

        # transform patch
        patch_t, mask_t = adv.patch_transformer(
            patch=patch,
            mask=mask,
            batch_size=1,
            img_size=min(args.width, args.height),
            do_rotate=True,
            rand_loc=True,
            random_size=True,
            train=False
        )

        if args.no_overwrite:
            attacked_img = img
        else:
            # apply transformed patch to clean image
            attacked_img = adv.patch_applier(img, patch_t, mask_t)

        if 'distill' in args.model:
            est_disp = models.distill(attacked_img)

            attacked_img = np.transpose(attacked_img[0].data.cpu().numpy(), (1, 2, 0))
            original_img = np.transpose(img[0].data.cpu().numpy(), (1, 2, 0))
            attacked_disp = est_disp[0, 0].data.cpu().numpy()
            original_disp = original_disp[0, 0].data.cpu().numpy()

            cv2.imwrite(save_path+'/img_attacked.png', attacked_img*255)
            plt.imsave(save_path+'/disp_attacked.png', attacked_disp, cmap='plasma', vmin=0, vmax=40)
            if not args.no_overwrite:
                cv2.imwrite(save_path+'/img_original.png', original_img*255)
                plt.imsave(save_path+'/disp_original.png', original_disp, cmap='plasma', vmin=0, vmax=40)
                plt.imsave(save_path+'/difference.png', abs(attacked_disp-original_disp), cmap='bwr')

    print("Finish test !")


if __name__ == '__main__':
    main()
