import torch
import torch.nn as nn
import numpy as np
import cv2
import math
import torch.nn.functional as F
import torchgeometry as tgm


class CreateFakeDisp(nn.Module):
    def __init__(self):
        super(CreateFakeDisp, self).__init__()

    def forward(self, original_disp, mask, target_disp):
        mask = mask[0, :, :]
        fake_disp = torch.mul((1 - mask), original_disp) + torch.mul(mask, target_disp)

        return fake_disp


class PatchApplier(nn.Module):
    def __init__(self):
        super(PatchApplier, self).__init__()

    def forward(self, img_batch, patch, mask):
        patched_img_batch = torch.mul((1 - mask), img_batch) + torch.mul(mask, patch)
        return patched_img_batch


class PatchTransformer(nn.Module):
    def __init__(self):
        super(PatchTransformer, self).__init__()
        self.min_contrast = 0.9
        self.max_contrast = 1.1
        self.min_brightness = -0.05
        self.max_brightness = 0.05
        self.noise_factor = 0.10
        self.minangle = -20
        self.maxangle = 20
        self.minsize = 0.35
        self.maxsize = 0.45

        self.min_x_off = -200
        self.max_x_off = 200
        self.min_y_off = -80
        self.max_y_off = 80
        self.max_x_trans = 0.1
        self.min_x_trans = -0.1
        self.max_y_trans = 0.1
        self.min_y_trans = -0.1

    def normalize_transforms(self, transforms, W, H):
        theta = torch.zeros(transforms.shape[0], 2, 3).cuda()
        theta[:, 0, 0] = transforms[:, 0, 0]
        theta[:, 0, 1] = transforms[:, 0, 1]*H/W
        theta[:, 0, 2] = transforms[:, 0, 2]*2/W + theta[:, 0, 0] + theta[:, 0, 1] - 1

        theta[:, 1, 0] = transforms[:, 1, 0]*W/H
        theta[:, 1, 1] = transforms[:, 1, 1]
        theta[:, 1, 2] = transforms[:, 1, 2]*2/H + theta[:, 1, 0] + theta[:, 1, 1] - 1

        return theta

    def forward(self, patch, mask, batch_size, img_size, do_rotate=True, rand_loc=True, random_size=True, do_perspective=True, train=True):
        # Determine size of padding
        pad = (img_size - patch.size(-1)) / 2

        # Make a batch of patches
        adv_batch = patch.expand(batch_size, -1, -1, -1)
        mask_batch = mask.expand(batch_size, -1, -1, -1)

        if train:
            # Contrast transforms
            contrast = torch.cuda.FloatTensor(adv_batch.size()[-3:]).uniform_(self.min_contrast, self.max_contrast).expand(batch_size, -1, -1, -1)

            # Brightness transforms
            brightness = torch.cuda.FloatTensor(adv_batch.size()[-3:]).uniform_(self.min_brightness, self.max_brightness).expand(batch_size, -1, -1, -1)

            # Noise transforms
            noise = torch.cuda.FloatTensor(adv_batch.size()[-3:]).uniform_(-1, 1).expand(batch_size, -1, -1, -1) * self.noise_factor

            # Apply contrast/brightness/noise transformation, and then clamp
            adv_batch = adv_batch * contrast + brightness + noise
            adv_batch = torch.clamp(adv_batch, 0.000001, 0.99999)

        # Pad patch and mask to image dimensions
        mypad = nn.ConstantPad2d((128, 128, int(pad + 0.5), int(pad)), 0)
        adv_batch = mypad(adv_batch)
        mask_batch = mypad(mask_batch)

        # Rotation and rescaling transforms
        if do_rotate:
            angle = torch.cuda.FloatTensor([1]).uniform_(self.minangle, self.maxangle).expand(batch_size)
        else:
            angle = torch.cuda.FloatTensor([1]).fill_(0).expand(batch_size)

        # Resize
        current_patch_size = adv_batch.size(-2)
        if random_size:
            size = torch.cuda.FloatTensor([1]).uniform_(self.minsize, self.maxsize)
            target_size = current_patch_size * (size ** 2)
        else:
            target_size = torch.Tensor([current_patch_size * (0.4 ** 2)]).cuda()
        scale = target_size / current_patch_size
        scale_b = scale.expand(batch_size)

        # Rotate
        center = (adv_batch.shape[3] / 2, adv_batch.shape[2] / 2)
        center = torch.Tensor(center).cuda()
        center = center.expand(batch_size, -1)

        rotation = tgm.get_rotation_matrix2d(center, angle, scale_b)
        rotation = nn.ConstantPad2d((0, 0, 0, 1), 0)(rotation)
        rotation[:, -1, -1] = 1.

        # Translation
        translation = torch.eye(3, 3).cuda()
        translation = translation.expand(batch_size, -1, -1)

        if rand_loc:
            x_off = torch.cuda.FloatTensor([1]).uniform_(self.min_x_off, self.max_x_off) / scale
            y_off = torch.cuda.FloatTensor([1]).uniform_(self.min_y_off, self.max_y_off) / scale
            translation[0, 0, 2] = x_off
            translation[0, 1, 2] = y_off

        if do_perspective:
            # Perspective transform
            src_coord = torch.Tensor([[[0, 0], [1, 0], [0, 1], [1, 1]]]).cuda()
            a_x = torch.cuda.FloatTensor([1]).uniform_(self.min_x_trans, self.max_x_trans)
            a_y = torch.cuda.FloatTensor([1]).uniform_(self.min_y_trans, self.max_y_trans)
            b_x = torch.cuda.FloatTensor([1]).uniform_(self.min_x_trans, self.max_x_trans)
            b_y = torch.cuda.FloatTensor([1]).uniform_(self.min_y_trans, self.max_y_trans)
            c_x = torch.cuda.FloatTensor([1]).uniform_(self.min_x_trans, self.max_x_trans)
            c_y = torch.cuda.FloatTensor([1]).uniform_(self.min_y_trans, self.max_y_trans)
            d_x = torch.cuda.FloatTensor([1]).uniform_(self.min_x_trans, self.max_x_trans)
            d_y = torch.cuda.FloatTensor([1]).uniform_(self.min_y_trans, self.max_y_trans)
            dst_coord = torch.Tensor([[[0+a_x, 0+a_y], [1+b_x, 0+b_y], [0+c_x, 1+c_y], [1+d_x, 1+d_y]]]).cuda()

            perspective = tgm.get_perspective_transform(src_coord, dst_coord).expand(batch_size, -1, -1)

            M = torch.bmm(torch.bmm(rotation, translation), perspective)
        else:
            M = torch.bmm(rotation, translation)

        M = torch.inverse(M)
        M = self.normalize_transforms(M, 512, 256)

        grid = F.affine_grid(M, adv_batch.shape)
        adv_batch_t = F.grid_sample(adv_batch, grid)
        mask_batch_t = F.grid_sample(mask_batch, grid)

        return adv_batch_t, mask_batch_t


class NonPrintabilityScore(nn.Module):
    def __init__(self, printability_file, patch_side):
        super(NonPrintabilityScore, self).__init__()
        self.printability_array = nn.Parameter(self.get_printability_array(printability_file, patch_side), requires_grad=False)

    def forward(self, patch):
        # calculate euclidian distance between colors in patch and colors in printability_array
        # square root of sum of squared difference
        color_dist = (patch - self.printability_array+0.000001)
        color_dist = color_dist ** 2
        color_dist = torch.sum(color_dist, 1)+0.000001
        color_dist = torch.sqrt(color_dist)
        # only work with the min distance
        color_dist_prod = torch.min(color_dist, 0)[0]
        # calculate the nps by summing over all pixels
        nps_score = torch.sum(color_dist_prod, 0)
        nps_score = torch.sum(nps_score, 0)
        return nps_score/torch.numel(patch)

    def get_printability_array(self, printability_file, side):
        printability_list = []
        with open(printability_file) as f:
            for line in f:
                printability_list.append(line.split(","))

        printability_array = []
        for printability_triplet in printability_list:
            printability_imgs = []
            red, green, blue = printability_triplet
            printability_imgs.append(np.full((side, side), red))
            printability_imgs.append(np.full((side, side), green))
            printability_imgs.append(np.full((side, side), blue))
            printability_array.append(printability_imgs)

        printability_array = np.asarray(printability_array)
        printability_array = np.float32(printability_array)
        pa = torch.from_numpy(printability_array)
        return pa


class TotalVariation(nn.Module):
    def __init__(self):
        super(TotalVariation, self).__init__()

    def forward(self, patch):
        tvcomp1 = torch.sum(torch.abs(patch[:, :, 1:] - patch[:, :, :-1]+0.000001), 0)
        tvcomp1 = torch.sum(torch.sum(tvcomp1, 0), 0)
        tvcomp2 = torch.sum(torch.abs(patch[:, 1:, :] - patch[:, :-1, :]+0.000001), 0)
        tvcomp2 = torch.sum(torch.sum(tvcomp2, 0), 0)
        tv = tvcomp1 + tvcomp2
        return tv/torch.numel(patch)


class PatchFunction(object):
    def __init__(self):
        super(PatchFunction, self).__init__()

    def LoadPatchFromImage(self, image_path, mask_path):
        noise_size = np.floor(self.image_size * np.sqrt(self.patch_size))
        patch_image = np.array(cv2.imread(image_path)).astype(np.float32)
        patch_image = self.resize_img(patch_image, (int(noise_size), int(noise_size)))/128. - 1
        patch = torch.FloatTensor(np.array([patch_image.transpose(2, 0, 1)]))
        mask_image = np.array(cv2.imread(mask_path)).astype(np.float32)
        mask_image = self.resize_img(mask_image, (int(noise_size), int(noise_size)))/256.
        mask = torch.FloatTensor(np.array([mask_image.transpose(2, 0, 1)]))
        return patch, mask

    def LoadPatchFromNpy(self, image_path, mask_path):
        patch_image = np.load(image_path).astype(np.float32)
        mask_image = np.load(mask_path).astype(np.float32)
        patch = torch.FloatTensor(patch_image)
        mask = torch.FloatTensor(mask_image)
        return patch, mask

    def InitSquarePatch(self, image_size, patch_size):
        noise_dim = patch_size
        patch = torch.rand((3, noise_dim, noise_dim))
        mask = torch.full((3, noise_dim, noise_dim), 1, dtype=torch.float32)
        return patch, mask

    def InitCirclePatch(self, image_size, patch_size):
        patch, _ = self.InitSquarePatch(image_size, patch_size)
        mask = self.CreateCircularMask(patch.shape[-2], patch.shape[-1]).astype('float32')
        mask = torch.FloatTensor(np.array([mask, mask, mask]))
        return patch, mask

    def CreateCircularMask(self, w, h):
        center = [int(w/2), int(h/2)]
        radius = min(center[0], center[1], w-center[0], h-center[1])-2
        Y, X = np.ogrid[:h, :w]
        dist_from_center = np.sqrt((X - center[0])**2 + (Y-center[1])**2)
        mask = dist_from_center <= radius
        return mask

    def resize_img(self, img, size):
        width, height = size
        return cv2.resize(img, (width, height), interpolation=cv2.INTER_AREA)


class Adversarial_Patch(PatchFunction):
    def __init__(self, patch_type, batch_size, image_size, patch_size, train=True, printability_file=None):
        super(Adversarial_Patch, self).__init__()
        self.patch_type = patch_type
        self.image_size = image_size
        self.patch_size = patch_size
        self.patch_transformer = PatchTransformer()
        self.patch_applier = PatchApplier()

        if train:
            self.printfile = printability_file
            self.create_fake_disp = CreateFakeDisp()
            self.calculate_nps = NonPrintabilityScore(self.printfile, self.patch_size).cuda()
            self.calculate_tv = TotalVariation().cuda()

    def initialize_patch_and_mask(self):
        if self.patch_type == 'square':
            patch, mask = self.InitSquarePatch(self.image_size, self.patch_size)
        elif self.patch_type == 'circle':
            patch, mask = self.InitCirclePatch(self.image_size, self.patch_size)
        print('=> Use "{}" patch with shape {}'.format(self.patch_type, patch.shape))
        return patch, mask

    def load_patch_and_mask_from_file(self, patch_path, mask_path, npy=True):
        patch, mask = self.LoadPatchFromNpy(patch_path, mask_path) if npy else self.LoadPatchFromImage(patch_path, mask_path)
        print('=> Load patch with shape {} from \'{}\''.format(patch.shape, patch_path))
        return patch, mask
